@@ -1,5 +1,5 @@
 /*
- *	$Id: ipfwadm.c,v 1.5 1995/04/19 20:28:49 jos Exp $
+ *	$Id: ipfwadm.c,v 1.11 1996/05/05 17:08:50 jos Exp $
  *
  *
  *	ipfwadm -- IP firewall and accounting administration
@@ -8,7 +8,7 @@
  *	about proper usage of this program.
  *
  *
- *	Copyright (C) 1995 by X/OS Experts in Open Systems BV.
+ *	Copyright (c) 1995,1996 by X/OS Experts in Open Systems BV.
  *	All rights reserved.
  *
  *	Author: Jos Vos <jos@xos.nl>
@@ -46,6 +46,23 @@
  *	1.4	Change the usage messages and add a large help text for -h.
  *	1.5	Change code to make "gcc -Wall" happy (and a few of these
  *		warnings were really serious ...).
+ *	1.6	Port to 1.3.x Linux kernel (1.3.57 as reference).
+ *		Rename -k option to -o (kernel packet logging).
+ *		Support new TCPACK option with new -k option.
+ *		Add support for masquerading rules (policy added).
+ *		List active masquerading entries (via -M flag).
+ *	1.7	Add some extra command/option checks and fix some typos.
+ *		Allow specification of ICMP types (given as port numbers) and
+ *		show source port numbers (ICMP types) when listing ICMP rules.
+ *		Eliminate warnings due to new 1.3.x. IP-related include files.
+ *	1.8	Major changes:
+ *		Support separate input/output chains, socket option renaming,
+ *		new -t flag for specifying TOS masks, and more...
+ *		Works (only!) on Linux 1.3.61 and higher.
+ *	1.9	Support matching with interface names (using option -W).
+ *	1.10	Change layout of listing of active masqueraded sessions.
+ *		Port to AXP (64-bit issues), untested.
+ *	1.11	Fix bugs introduced in 1.10.
  */
 
 #include <stdio.h>
@@ -58,25 +75,30 @@
 #include <linux/ip.h>
 #include <linux/tcp.h>
 #include <linux/udp.h>
+#include <linux/icmp.h>
+#include <linux/if.h>
 #include <linux/ip_fw.h>
 
 #define IP_VERSION	4
 #define IP_OFFSET	0x1FFF
 
 #define CHN_NONE	-1
-#define CHN_BLOCK	0
-#define CHN_FORWARD	1
-#define CHN_ACCT	2
+#define CHN_FWD		0
+#define CHN_IN		1
+#define CHN_OUT		2
+#define CHN_ACCT	3
+#define CHN_MASQ	4	/* only used for listing masquerading */
 
 #define CMD_NONE	0x0000
 #define CMD_LIST	0x0001
-#define CMD_ADD		0x0002
+#define CMD_APPEND	0x0002
 #define CMD_DELETE	0x0004
 #define CMD_FLUSH	0x0008
 #define CMD_RESET	0x0010
 #define CMD_POLICY	0x0020
 #define CMD_CHECK	0x0040
 #define CMD_HELP	0x0080
+#define CMD_INSERT	0x0100
 
 #define OPT_NONE	0x0000
 #define OPT_EXTENDED	0x0001
@@ -84,13 +106,16 @@
 #define OPT_SOURCE	0x0004
 #define OPT_DESTINATION	0x0008
 #define OPT_PROTOCOL	0x0010
-#define OPT_VIAHOST	0x0020
+#define OPT_VIAADDR	0x0020
 #define OPT_POLICY	0x0040
 #define OPT_TCPSYN	0x0080
 #define OPT_BIDIR	0x0100
 #define OPT_VERBOSE	0x0200
 #define OPT_PRINTK	0x0400
 #define OPT_EXPANDED	0x0800
+#define OPT_TCPACK	0x1000
+#define OPT_TOS		0x2000
+#define OPT_VIANAME	0x4000
 
 #define FMT_NUMERIC	0x0001
 #define FMT_NOCOUNTS	0x0002
@@ -99,19 +124,41 @@
 #define FMT_NOTABLE	0x0010
 #define FMT_HEADER	0x0020
 #define FMT_NOKIND	0x0040
-#define FMT_VIAHOST	0x0080
+#define FMT_VIA		0x0080
 #define FMT_NONEWLINE	0x0100
+#define FMT_DELTAS	0x0200
+#define FMT_TOS		0x0400
 
-char ipfwadm_version[] = "$Id: ipfwadm.c,v 1.5 1995/04/19 20:28:49 jos Exp $";
+struct masq {
+	unsigned long	expires;	/* Expiration timer */
+	unsigned short	kind;		/* Which protocol are we talking? */
+	struct in_addr	src, dst;	/* Source and destination IP addresses */
+	unsigned short	sport, dport;	/* Source and destination ports */
+	unsigned short	mport;		/* Masqueraded port */
+	__u32		initseq;	/* Add delta from this seq. on */
+	short		delta;		/* Delta in sequence numbers */
+	short		pdelta;		/* Delta in sequence numbers before last */
+};
 
-int ssocmd_add[3] = {IP_FW_ADD_BLK, IP_FW_ADD_FWD, IP_ACCT_ADD};
-int ssocmd_delete[3] = {IP_FW_DEL_BLK, IP_FW_DEL_FWD, IP_ACCT_DEL};
-int ssocmd_check[3] = {IP_FW_CHK_BLK, IP_FW_CHK_FWD, 0};
-int ssocmd_policy[3] = {IP_FW_POLICY_BLK, IP_FW_POLICY_FWD, 0};
-int ssocmd_flush[3] = {IP_FW_FLUSH_BLK, IP_FW_FLUSH_FWD, IP_ACCT_FLUSH};
-int ssocmd_zero[3] = {IP_FW_ZERO_BLK, IP_FW_ZERO_FWD, IP_ACCT_ZERO};
-char *procfiles[3] = {"/proc/net/ip_block", "/proc/net/ip_forward",
-	"/proc/net/ip_acct"};
+char ipfwadm_version[] = "$Id: ipfwadm.c,v 1.11 1996/05/05 17:08:50 jos Exp $";
+char package_version[] = "ipfwadm 2.0, 1996/05/05";
+
+int ssocmd_insert[4] =
+	{IP_FW_INSERT_FWD, IP_FW_INSERT_IN, IP_FW_INSERT_OUT, IP_ACCT_INSERT};
+int ssocmd_append[4] =
+	{IP_FW_APPEND_FWD, IP_FW_APPEND_IN, IP_FW_APPEND_OUT, IP_ACCT_APPEND};
+int ssocmd_delete[4] =
+	{IP_FW_DELETE_FWD, IP_FW_DELETE_IN, IP_FW_DELETE_OUT, IP_ACCT_DELETE};
+int ssocmd_check[4] =
+	{IP_FW_CHECK_FWD, IP_FW_CHECK_IN, IP_FW_CHECK_OUT, -1};
+int ssocmd_policy[4] =
+	{IP_FW_POLICY_FWD, IP_FW_POLICY_IN, IP_FW_POLICY_OUT, -1};
+int ssocmd_flush[4] =
+	{IP_FW_FLUSH_FWD, IP_FW_FLUSH_IN, IP_FW_FLUSH_OUT, IP_ACCT_FLUSH};
+int ssocmd_zero[4] =
+	{IP_FW_ZERO_FWD, IP_FW_ZERO_IN, IP_FW_ZERO_OUT, IP_ACCT_ZERO};
+char *procfiles[5] = {"/proc/net/ip_forward", "/proc/net/ip_input",
+	"/proc/net/ip_output", "/proc/net/ip_acct", "/proc/net/ip_masquerade"};
 
 int chain = CHN_NONE;
 int command = CMD_NONE;
@@ -137,6 +184,7 @@ extern void parse_protocol(char *);
 extern void parse_all_ports(char **, unsigned short *, int, int);
 extern unsigned short parse_port(char *, unsigned short);
 extern void store_port(char *, unsigned short *, int, char *[]);
+extern void parse_hexbyte(char *, unsigned char *);
 
 extern struct in_addr *host_to_addr(char *, int *);
 extern char *addr_to_host(struct in_addr *);
@@ -154,9 +202,12 @@ extern char *policy_to_string(int);
 extern int add_delete_entries(int, int);
 extern int check_entries(int);
 extern int list_entries(int, char *);
+extern int list_masq();
 
 extern void print_firewall(FILE *, struct ip_fw *, int);
+extern void print_masq(FILE *, struct masq *, int);
 extern int read_procinfo(FILE *, struct ip_fw *, int);
+extern int read_masqinfo(FILE *, struct masq *, int);
 extern struct ip_fwpkt *fw_to_fwpkt(struct ip_fw *);
 extern int do_setsockopt(int, void *, int);
 
@@ -176,27 +227,37 @@ main(int argc, char *argv[])
 
 	program = argv[0];
 
-	while ((c = getopt(argc, argv, "ABFadlzfp:chP:S:D:I:beknvxy")) != -1)
+	while ((c = getopt(argc, argv, "AFIMOadilzfp:chP:S:D:V:W:beknot:vxy")) != -1)
 		switch (c) {
 		case 'A':
 			if (chain != CHN_NONE)
 				exit_error(2, "multiple categories specified");
 			chain = CHN_ACCT;
 			break;
-		case 'B':
-			if (chain != CHN_NONE)
-				exit_error(2, "multiple categories specified");
-			chain = CHN_BLOCK;
-			break;
 		case 'F':
 			if (chain != CHN_NONE)
 				exit_error(2, "multiple categories specified");
-			chain = CHN_FORWARD;
+			chain = CHN_FWD;
+			break;
+		case 'I':
+			if (chain != CHN_NONE)
+				exit_error(2, "multiple categories specified");
+			chain = CHN_IN;
+			break;
+		case 'M':
+			if (chain != CHN_NONE)
+				exit_error(2, "multiple categories specified");
+			chain = CHN_MASQ;
+			break;
+		case 'O':
+			if (chain != CHN_NONE)
+				exit_error(2, "multiple categories specified");
+			chain = CHN_OUT;
 			break;
 		case 'a':
 			if (command != CMD_NONE)
 				exit_error(2, "multiple commands specified");
-			command = CMD_ADD;
+			command = CMD_APPEND;
 			if (optind < argc && argv[optind][0] != '-') {
 				parse_policy(argv[optind++]);
 				options |= OPT_POLICY;
@@ -211,6 +272,15 @@ main(int argc, char *argv[])
 				options |= OPT_POLICY;
 			}
 			break;
+		case 'i':
+			if (command != CMD_NONE)
+				exit_error(2, "multiple commands specified");
+			command = CMD_INSERT;
+			if (optind < argc && argv[optind][0] != '-') {
+				parse_policy(argv[optind++]);
+				options |= OPT_POLICY;
+			}
+			break;
 		case 'l':
 			if (command != CMD_NONE && command != CMD_RESET)
 				exit_error(2, "multiple commands specified");
@@ -244,7 +314,6 @@ main(int argc, char *argv[])
 		case 'P':
 			check_option(OPT_PROTOCOL, 'P');
 			parse_protocol(optarg);
-			options |= OPT_PROTOCOL;
 			break;
 		case 'S':
 			check_option(OPT_SOURCE, 'S');
@@ -252,7 +321,6 @@ main(int argc, char *argv[])
 			while (optind < argc && argv[optind][0] != '-')
 				store_port(argv[optind++], &firewall.fw_nsp,
 					IP_FW_F_SRNG, sports);
-			options |= OPT_SOURCE;
 			break;
 		case 'D':
 			check_option(OPT_DESTINATION, 'D');
@@ -260,42 +328,49 @@ main(int argc, char *argv[])
 			while (optind < argc && argv[optind][0] != '-')
 				store_port(argv[optind++], &firewall.fw_ndp,
 					IP_FW_F_DRNG, dports);
-			options |= OPT_DESTINATION;
 			break;
-		case 'I':
-			check_option(OPT_VIAHOST, 'I');
+		case 'V':
+			check_option(OPT_VIAADDR, 'V');
 			parse_viahost(optarg);
-			options |= OPT_VIAHOST;
+			break;
+		case 'W':
+			check_option(OPT_VIANAME, 'W');
+			strncpy(firewall.fw_vianame, optarg, IFNAMSIZ);
 			break;
 		case 'b':
 			check_option(OPT_BIDIR, 'b');
-			options |= OPT_BIDIR;
 			firewall.fw_flg |= IP_FW_F_BIDIR;
 			break;
 		case 'e':
 			check_option(OPT_EXTENDED, 'e');
-			options |= OPT_EXTENDED;
 			break;
 		case 'k':
-			check_option(OPT_PRINTK, 'k');
-			options |= OPT_PRINTK;
-			firewall.fw_flg |= IP_FW_F_PRN;
+			check_option(OPT_TCPACK, 'k');
+			firewall.fw_flg |= IP_FW_F_TCPACK;
 			break;
 		case 'n':
 			check_option(OPT_NUMERIC, 'n');
-			options |= OPT_NUMERIC;
+			break;
+		case 'o':
+			check_option(OPT_PRINTK, 'o');
+			firewall.fw_flg |= IP_FW_F_PRN;
+			break;
+		case 't':
+			check_option(OPT_TOS, 't');
+			if (optind < argc && argv[optind][0] != '-') {
+				parse_hexbyte(optarg, &firewall.fw_tosand);
+				parse_hexbyte(argv[optind++], &firewall.fw_tosxor);
+			} else
+				exit_error(2, "-t requires 2 hexbyte arguments");
 			break;
 		case 'v':
 			check_option(OPT_VERBOSE, 'v');
-			options |= OPT_VERBOSE;
 			break;
 		case 'x':
 			check_option(OPT_EXPANDED, 'x');
-			options |= OPT_EXPANDED;
 			break;
 		case 'y':
 			check_option(OPT_TCPSYN, 'y');
-			options |= OPT_TCPSYN;
 			firewall.fw_flg |= IP_FW_F_TCPSYN;
 			break;
 		case '?':
@@ -306,54 +381,81 @@ main(int argc, char *argv[])
 	if (optind < argc)
 		exit_error(2, "unknown arguments found on commandline");
 
+	kind = firewall.fw_flg & IP_FW_F_KIND;
+	policy = (int) (firewall.fw_flg &
+		(IP_FW_F_ACCEPT | IP_FW_F_ICMPRPL | IP_FW_F_MASQ));
+
 	if (chain == CHN_NONE)
-		exit_error(2, "missing -A, -B, or -F flag");
-	else if (chain != CHN_BLOCK && chain != CHN_FORWARD &&
+		exit_error(2, "missing -A, -F, -I, -O, or -M flag");
+	else if (chain != CHN_IN && chain != CHN_OUT && chain != CHN_FWD &&
 			command == CMD_CHECK)
-		exit_error(2, "specify either -B or -F for checking a packet");
-	else if (chain != CHN_BLOCK && chain != CHN_FORWARD &&
+		exit_error(2, "specify either -F, -I, or -O for checking a packet");
+	else if (chain != CHN_IN && chain != CHN_OUT && chain != CHN_FWD &&
 			command == CMD_POLICY)
-		exit_error(2, "specify either -B or -F for changing the policy");
-	else if ((chain == CHN_BLOCK || chain == CHN_FORWARD) &&
-			!(options & OPT_POLICY) && (command & (CMD_ADD | CMD_DELETE)))
+		exit_error(2, "specify either -F, -I, or -O for changing the policy");
+	else if ((chain == CHN_IN || chain == CHN_OUT || chain == CHN_FWD) &&
+			!(options & OPT_POLICY) &&
+			(command & (CMD_INSERT | CMD_APPEND | CMD_DELETE)))
 		exit_error(2, "policy required for firewall entries");
-	else if (chain != CHN_BLOCK && chain != CHN_FORWARD &&
+	else if (chain != CHN_IN && chain != CHN_OUT && chain != CHN_FWD &&
 			options & OPT_POLICY)
-		/* command is CMD_ADD or CMD_DELETE in this case */
+		/* command is CMD_INSERT, CMD_APPEND, or CMD_DELETE */
 		exit_error(2, "no policy allowed with non-firewall entries");
+	else if (chain == CHN_MASQ && !(command & CMD_LIST))
+		exit_error(2, "-M only allowed in combination with -l command");
 
-	if ((options & OPT_BIDIR) && !(command & (CMD_ADD | CMD_DELETE)))
-		exit_error(2, "bidirectional flag (-b) only allowed with add/delete");
-
-	if ((options & OPT_PRINTK) && !(command & (CMD_ADD | CMD_DELETE)))
-		exit_error(2, "kernel print flag (-k) only allowed with add/delete");
+	if ((options & OPT_BIDIR) &&
+			!(command & (CMD_INSERT | CMD_APPEND | CMD_DELETE)))
+		exit_error(2, "bidirectional flag (-b) only allowed with insert/append/delete");
+	else if ((options & OPT_PRINTK) &&
+			!(command & (CMD_INSERT | CMD_APPEND | CMD_DELETE)))
+		exit_error(2, "kernel print flag (-o) only allowed with insert/append/delete");
+	else if ((options & OPT_TOS) &&
+			!(command & (CMD_INSERT | CMD_APPEND | CMD_DELETE)))
+		exit_error(2, "tos values (-t) only allowed with insert/append/delete");
 
 	if ((options & OPT_PROTOCOL) && (command & (CMD_LIST | CMD_FLUSH |
 			CMD_RESET | CMD_POLICY)))
-		exit_error(2, "no protocol(-P) allowed with this command");
+		exit_error(2, "no protocol (-P) allowed with this command");
 	else if (!(options & OPT_PROTOCOL) && command == CMD_CHECK)
 		exit_error(2, "protocol (-P) required for this command");
 	else if (!(options & OPT_PROTOCOL))
 		firewall.fw_flg |= IP_FW_F_ALL;
 
-	kind = firewall.fw_flg & IP_FW_F_KIND;
 	if (command == CMD_CHECK && kind == IP_FW_F_ALL)
 		exit_error(2, "specific protocol required for check command");
 
 	if (firewall.fw_flg & IP_FW_F_TCPSYN && kind != IP_FW_F_TCP)
-		exit_error(2, "SYN option (-s) only allowed for TCP protocol");
+		exit_error(2, "SYN option (-y) only allowed for TCP protocol");
+	else if (firewall.fw_flg & IP_FW_F_TCPACK && kind != IP_FW_F_TCP)
+		exit_error(2, "ACK option (-k) only allowed for TCP protocol");
 
-	if ((options & OPT_VIAHOST) && (command & (CMD_LIST | CMD_FLUSH |
+	if ((options & OPT_VIAADDR) && (command & (CMD_LIST | CMD_FLUSH |
 			CMD_RESET | CMD_POLICY)))
-		exit_error(2, "no interface (-I) allowed with this command");
-	if (!(options & OPT_VIAHOST) && (command & CMD_CHECK))
-		exit_error(2, "interface (-I) required for this command");
+		exit_error(2, "no interface (-V) allowed with this command");
+	if (!(options & OPT_VIAADDR) && (command & CMD_CHECK))
+		exit_error(2, "interface (-V) required for this command");
+
+	if (kind == IP_FW_F_ICMP && firewall.fw_ndp != 0)
+		exit_error(2, "ICMP types only allowed with -S, not with -D");
+	else if (kind == IP_FW_F_ALL && (firewall.fw_nsp != 0 || firewall.fw_ndp != 0))
+		exit_error(2, "no ports allowed without specific protocol");
+	else if (command == CMD_CHECK && kind != IP_FW_F_ICMP &&
+			(firewall.fw_nsp != 1 || firewall.fw_ndp != 1))
+		exit_error(2, "one port required with source/destination address");
+	else if (command == CMD_CHECK && kind == IP_FW_F_ICMP &&
+			(firewall.fw_nsp != 1 || firewall.fw_ndp != 0))
+		exit_error(2, "ICMP type required after source address");
+	else if ((policy & IP_FW_F_MASQ) && chain != CHN_FWD)
+		exit_error(2, "masquerading only allowed in combination with -F");
+	else if (kind == IP_FW_F_ICMP && (policy & IP_FW_F_MASQ))
+		exit_error(2, "masquerading not allowed with protocol ICMP");
 
 	if ((options & OPT_SOURCE) && (command & (CMD_LIST | CMD_FLUSH |
 			CMD_RESET | CMD_POLICY)))
 		exit_error(2, "no source address (-S) allowed with this command");
-	else if (!(options & OPT_SOURCE) && (command & (CMD_ADD | CMD_DELETE |
-			CMD_CHECK)))
+	else if (!(options & OPT_SOURCE) && (command &
+			(CMD_INSERT | CMD_APPEND | CMD_DELETE | CMD_CHECK)))
 		exit_error(2, "source address (-S) required for this command");
 	else if (options & OPT_SOURCE) {
 		parse_hostnetworkmask(shostnetworkmask, &saddrs,
@@ -366,7 +468,7 @@ main(int argc, char *argv[])
 			CMD_RESET | CMD_POLICY)))
 		exit_error(2, "no destination address (-D) allowed with this command");
 	else if (!(options & OPT_DESTINATION) && (command &
-			(CMD_ADD | CMD_DELETE | CMD_CHECK)))
+			(CMD_INSERT | CMD_APPEND | CMD_DELETE | CMD_CHECK)))
 		exit_error(2, "destination address (-D) required for this command");
 	else if (options & OPT_DESTINATION) {
 		parse_hostnetworkmask(dhostnetworkmask, &daddrs,
@@ -375,17 +477,17 @@ main(int argc, char *argv[])
 			(int) firewall.fw_ndp, firewall.fw_flg & IP_FW_F_DRNG);
 	}
 
-	if (kind == IP_FW_F_ICMP && (firewall.fw_nsp != 0 || firewall.fw_ndp != 0))
-		exit_error(2, "no ports allowed with protocol ICMP");
-	if (kind == IP_FW_F_ALL && (firewall.fw_nsp != 0 || firewall.fw_ndp != 0))
-		exit_error(2, "no ports allowed without specific protocol");
-	else if (command == CMD_CHECK && kind != IP_FW_F_ICMP &&
-			(firewall.fw_nsp != 1 || firewall.fw_ndp != 1))
-		exit_error(2, "one port required with source/destination address");
+	if (!(options & OPT_TOS)) {
+		firewall.fw_tosand = 0xFF;
+		firewall.fw_tosxor = 0x00;
+	}
 
 	switch (command) {
-	case CMD_ADD:
-		ret = add_delete_entries(ssocmd_add[chain], chain);
+	case CMD_INSERT:
+		ret = add_delete_entries(ssocmd_insert[chain], chain);
+		break;
+	case CMD_APPEND:
+		ret = add_delete_entries(ssocmd_append[chain], chain);
 		break;
 	case CMD_DELETE:
 		ret = add_delete_entries(ssocmd_delete[chain], chain);
@@ -394,7 +496,6 @@ main(int argc, char *argv[])
 		ret = check_entries(ssocmd_check[chain]);
 		break;
 	case CMD_POLICY:
-		policy = (int) (firewall.fw_flg & (IP_FW_F_ACCEPT | IP_FW_F_ICMPRPL));
 		ret = do_setsockopt(ssocmd_policy[chain], &policy, sizeof(int));
 		break;
 	case CMD_FLUSH:
@@ -405,7 +506,10 @@ main(int argc, char *argv[])
 		break;
 	case CMD_LIST:
 	case CMD_LIST | CMD_RESET:
-		ret = list_entries(chain, (command & CMD_RESET) ? "r+" : "r");
+		if (chain == CHN_MASQ)
+			ret = list_masq();
+		else
+			ret = list_entries(chain, (command & CMD_RESET) ? "r+" : "r");
 		break;
 	default:
 		/* We should never reach this... */
@@ -541,6 +645,8 @@ parse_policy(char *s)
 		policy = IP_FW_F_ACCEPT;
 	else if (strncmp("deny", s, strlen(s)) == 0)
 		policy = 0; /* as opposed to IP_FW_F_ACCEPT */
+	else if (strncmp("masquerade", s, strlen(s)) == 0)
+		policy = IP_FW_F_ACCEPT | IP_FW_F_MASQ;
 	else if (strncmp("reject", s, strlen(s)) == 0)
 		policy = IP_FW_F_ICMPRPL;
 	else {
@@ -592,7 +698,7 @@ parse_all_ports(char **ports, unsigned short *portnumbers,
 			portnumbers[0] = parse_port(buf, kind);
 			portnumbers[1] = parse_port(cp + 1, kind);
 			if (portnumbers[0] > portnumbers[1]) {
-				fprintf(stderr, "%s: invalid range of ports (%d > %d)\n",
+				fprintf(stderr, "%s: invalid range of ports (%u > %u)\n",
 					program, portnumbers[0], portnumbers[1]);
 				exit(2);
 			}
@@ -607,7 +713,12 @@ parse_port(char *port, unsigned short kind)
 
 	if ((portnum = string_to_number(port, 0, 65535)) != -1)
 		return (unsigned short) portnum;
-	else if ((portnum = service_to_port(port, kind)) != -1)
+	else if (kind == IP_FW_F_ICMP) {
+		/* ICMP types (given as port numbers) must be numeric! */
+		fprintf(stderr, "%s: invalid ICMP type \"%s\" specified\n",
+			program, port);
+		exit(2);
+	} else if ((portnum = service_to_port(port, kind)) != -1)
 		return (unsigned short) portnum;
 	else {
 		fprintf(stderr, "%s: invalid port/service \"%s\" specified\n",
@@ -630,7 +741,7 @@ store_port(char *port, unsigned short *nports, int rangeflag, char *ports[])
 	}
 
 	if (firewall.fw_nsp + firewall.fw_ndp > IP_FW_MAX_PORTS) {
-		fprintf(stderr, "%s: too many ports specified(maximum %d)\n",
+		fprintf(stderr, "%s: too many ports specified (maximum %d)\n",
 			program, IP_FW_MAX_PORTS);
 		exit(2);
 	}
@@ -638,6 +749,29 @@ store_port(char *port, unsigned short *nports, int rangeflag, char *ports[])
 	return;
 }
 
+void
+parse_hexbyte(char *s, unsigned char *b)
+{
+	long number;
+	char *end;
+
+	number = strtol(s, &end, 16);
+	if (*end == '\0' && end != s) {
+		/* we parsed a number, let's see if we want this */
+		if (0 <= number && number <= 255)
+			*b = (unsigned char) number;
+		else {
+			fprintf(stderr, "%s: invalid hexbyte \"%s\" specified\n",
+				program, s);
+			exit(2);
+		}
+	} else {
+		fprintf(stderr, "%s: invalid hexbyte \"%s\" specified\n",
+			program, s);
+		exit(2);
+	}
+}
+
 struct in_addr *
 host_to_addr(char *name, int *naddr)
 {
@@ -763,7 +897,7 @@ mask_to_dotted(struct in_addr *mask)
 {
 	int i;
 	static char buf[20];
-	unsigned long maskaddr, bits;
+	__u32 maskaddr, bits;
 
 	maskaddr = ntohl(mask->s_addr);
 
@@ -804,10 +938,10 @@ port_to_service(int port, unsigned short kind)
 	struct servent *service;
 
 	if (kind == IP_FW_F_TCP &&
-			(service = getservbyport(port, "tcp")) != NULL)
+			(service = getservbyport(htons(port), "tcp")) != NULL)
 		return service->s_name;
 	else if (kind == IP_FW_F_UDP &&
-			(service = getservbyport(port, "udp")) != NULL)
+			(service = getservbyport(htons(port), "udp")) != NULL)
 		return service->s_name;
 	else
 		return (char *) NULL;
@@ -836,6 +970,8 @@ policy_to_string(int policy)
 	switch (policy) {
 	case IP_FW_F_ACCEPT:
 		return "accept";
+	case IP_FW_F_ACCEPT | IP_FW_F_MASQ:
+		return "masquerade";
 	case IP_FW_F_ICMPRPL:
 		return "reject";
 	default:
@@ -853,9 +989,9 @@ add_delete_entries(int cmd, int chain)
 		for (j = 0; j < ndaddrs; j++) {
 			firewall.fw_dst.s_addr = daddrs[j].s_addr;
 			if (options & OPT_VERBOSE)
-				print_firewall(stdout, &firewall, FMT_NOCOUNTS | FMT_OPTIONS |
-					((chain == CHN_BLOCK || chain == CHN_FORWARD) ?
-					0 : FMT_NOKIND) | FMT_VIAHOST | FMT_NUMERIC | FMT_NOTABLE);
+				print_firewall(stdout, &firewall, FMT_NOCOUNTS | FMT_OPTIONS | FMT_TOS |
+					((chain == CHN_IN || chain == CHN_OUT|| chain == CHN_FWD) ?
+					0 : FMT_NOKIND) | FMT_VIA | FMT_NUMERIC | FMT_NOTABLE);
 			ret |= do_setsockopt(cmd, &firewall, sizeof(firewall));
 		}
 	}
@@ -874,7 +1010,7 @@ check_entries(int cmd)
 			firewall.fw_dst.s_addr = daddrs[j].s_addr;
 			if (options & OPT_VERBOSE)
 				print_firewall(stdout, &firewall, FMT_NOCOUNTS | FMT_NOKIND |
-					FMT_VIAHOST | FMT_NUMERIC | FMT_NOTABLE);
+					FMT_VIA | FMT_NUMERIC | FMT_NOTABLE);
 			packet = fw_to_fwpkt(&firewall);
 			ret |= do_setsockopt(cmd, packet, sizeof(struct ip_fwpkt));
 		}
@@ -899,7 +1035,7 @@ list_entries(int chain, char *mode)
 		exit(1);
 	}
 
-	if (chain == CHN_BLOCK || chain == CHN_FORWARD) {
+	if (chain == CHN_IN || chain == CHN_OUT || chain == CHN_FWD) {
 		if (fscanf(fp, "%[^,], default %d", buf, &policy) != 2) {
 			fprintf(stderr, "%s: unexpected input from %s\n",
 				program, procfile);
@@ -924,7 +1060,7 @@ list_entries(int chain, char *mode)
 	ntotal += nread;
 
 	format = 0;
-	if (chain == CHN_BLOCK || chain == CHN_FORWARD) {
+	if (chain == CHN_IN || chain == CHN_OUT || chain == CHN_FWD) {
 		if (!(options & OPT_EXTENDED))
 			format |= FMT_NOCOUNTS;
 	} else
@@ -933,8 +1069,10 @@ list_entries(int chain, char *mode)
 	if (options & OPT_NUMERIC)
 		format |= FMT_NUMERIC;
 
-	if (options & OPT_EXTENDED)
-		format |= FMT_OPTIONS | FMT_VIAHOST;
+	if (options & OPT_EXTENDED && chain != CHN_ACCT)
+		format |= FMT_OPTIONS | FMT_TOS | FMT_VIA;
+	else if (options & OPT_EXTENDED)
+		format |= FMT_OPTIONS | FMT_VIA;
 
 	if (!(options & OPT_EXPANDED))
 		format |= FMT_KILOMEGA;
@@ -947,6 +1085,55 @@ list_entries(int chain, char *mode)
 	return 0;
 }
 
+int
+list_masq()
+{
+	FILE *fp;
+	char *procfile;
+	int i;
+	char buf[256];
+	struct masq *mslist;
+	int ntotal = 0, nread, format;
+
+	procfile = procfiles[CHN_MASQ];
+
+	if ((fp = fopen(procfile, "r")) == NULL) {
+		fprintf(stderr, "%s: cannot open file %s\n", program, procfile);
+		exit(1);
+	}
+
+	if (fgets(buf, sizeof(buf), fp) == NULL) {
+		fprintf(stderr, "%s: unexpected input from %s\n",
+			program, procfile);
+		exit(2);
+	}
+
+	fputs("IP masquerading entries\n", stdout);
+
+	mslist = (struct masq *) fw_malloc(16 * sizeof(struct masq));
+	while ((nread = read_masqinfo(fp, &(mslist[ntotal]), 16)) == 16) {
+		ntotal += nread;
+		mslist = (struct masq *) fw_realloc(mslist,
+			(ntotal + 16) * sizeof(struct masq));
+	}
+	ntotal += nread;
+
+	format = 0;
+
+	if (options & OPT_NUMERIC)
+		format |= FMT_NUMERIC;
+
+	if (options & OPT_EXTENDED)
+		format |= FMT_DELTAS;
+
+	if (ntotal > 0)
+		for (i = 0; i < ntotal; i++)
+			print_masq(stdout, &(mslist[i]),
+				(i) ? format : (format | FMT_HEADER));
+
+	return 0;
+}
+
 void
 print_firewall(FILE *fp, struct ip_fw *fw, int format)
 {
@@ -975,9 +1162,13 @@ print_firewall(FILE *fp, struct ip_fw *fw, int format)
 			fputs("typ ", fp);
 		fputs("prot ", fp);
 		if (format & FMT_OPTIONS)
-			fputs("opt ", fp);
-		if (format & FMT_VIAHOST)
-			fprintf(fp, FMT("%-15s ","(%s) "), "interface");
+			fputs("opt  ", fp);
+		if (format & FMT_TOS)
+			fputs("tosa tosx ", fp);
+		if (format & FMT_VIA) {
+			fprintf(fp, FMT("%-7s ","(%s "), "ifname");
+			fprintf(fp, FMT("%-15s ","%s) "), "ifaddress");
+		}
 		fprintf(fp, FMT("%-20s ","%s "), "source");
 		fprintf(fp, FMT("%-20s ","%s "), "destination");
 		fputs("ports\n", fp);
@@ -1013,7 +1204,9 @@ print_firewall(FILE *fp, struct ip_fw *fw, int format)
 	}
 
 	if (!(format & FMT_NOKIND)) {
-		if (flags & IP_FW_F_ACCEPT)
+		if (flags & IP_FW_F_MASQ)
+			fprintf(fp, "msq ");
+		else if (flags & IP_FW_F_ACCEPT)
 			fprintf(fp, "acc ");
 		else if (flags & IP_FW_F_ICMPRPL)
 			fprintf(fp, "rej ");
@@ -1039,15 +1232,28 @@ print_firewall(FILE *fp, struct ip_fw *fw, int format)
 		if (format & FMT_NOTABLE)
 			fputs("opt ", fp);
 		fputc((flags & IP_FW_F_BIDIR) ? 'b' : '-', fp);
+		fputc((flags & IP_FW_F_TCPACK) ? 'k' : '-', fp);
 		fputc((flags & IP_FW_F_TCPSYN) ? 'y' : '-', fp);
-		fputc((flags & IP_FW_F_PRN) ? 'k' : '-', fp);
+		fputc((flags & IP_FW_F_PRN) ? 'o' : '-', fp);
 		fputc(' ', fp);
 	}
 
-	if (format & FMT_VIAHOST)
-		fprintf(fp, FMT("%-15s ","via %s "), (fw->fw_via.s_addr == 0L &&
+	if (format & FMT_TOS) {
+		if (format & FMT_NOTABLE)
+			fputs("tos ", fp);
+		fprintf(fp, "0x%02hX 0x%02hX ",
+			(unsigned short) fw->fw_tosand,
+			(unsigned short) fw->fw_tosxor);
+	}
+
+	if (format & FMT_VIA) {
+		fprintf(fp, FMT("%-7.16s ","via %.16s "),
+			(fw->fw_vianame)[0] ? fw->fw_vianame :
+				((format & FMT_NUMERIC) ? "*" : "any"));
+		fprintf(fp, FMT("%-15s ","%s "), (fw->fw_via.s_addr == 0L &&
 			!(format & FMT_NUMERIC)) ? "any" :
 			addr_to_dotted(&(fw->fw_via)));
+	}
 
 	if (format & FMT_NOTABLE)
 		fputs("  ", fp);
@@ -1077,7 +1283,7 @@ print_firewall(FILE *fp, struct ip_fw *fw, int format)
 	if (format & FMT_NOTABLE)
 		fputs("  ", fp);
 
-	if (kind != IP_FW_F_TCP && kind != IP_FW_F_UDP) {
+	if (kind != IP_FW_F_TCP && kind != IP_FW_F_UDP && kind != IP_FW_F_ICMP) {
 		fputs(" n/a", fp);
 		if (!(format & FMT_NONEWLINE))
 			putc('\n', fp);
@@ -1091,14 +1297,21 @@ print_firewall(FILE *fp, struct ip_fw *fw, int format)
 			fputc((i == 0) ? ' ' :
 				((flags & IP_FW_F_SRNG && i == 1) ? ':' : ','), fp);
 			if (format & FMT_NUMERIC)
-				fprintf(fp, "%d", fw->fw_pts[i]);
+				fprintf(fp, "%u", fw->fw_pts[i]);
 			else if ((service = port_to_service(fw->fw_pts[i],
 					kind)) != NULL)
 				fputs(service, fp);
 			else
-				fprintf(fp, "%d", fw->fw_pts[i]);
+				fprintf(fp, "%u", fw->fw_pts[i]);
 		}
 
+	/* only source ports are interpreted as ICMP types */
+	if (kind == IP_FW_F_ICMP) {
+		if (!(format & FMT_NONEWLINE))
+			putc('\n', fp);
+		return;
+	}
+
 	fputs(" ->", fp);
 
 	if (fw->fw_ndp == 0)
@@ -1108,45 +1321,146 @@ print_firewall(FILE *fp, struct ip_fw *fw, int format)
 			fputc((i == fw->fw_nsp) ? ' ' : ((flags & IP_FW_F_DRNG &&
 				i == (fw->fw_nsp + 1)) ? ':' : ','), fp);
 			if (format & FMT_NUMERIC)
-				fprintf(fp, "%d", fw->fw_pts[i]);
+				fprintf(fp, "%u", fw->fw_pts[i]);
 			else if ((service = port_to_service(fw->fw_pts[i], kind)) != NULL)
 				fputs(service, fp);
 			else
-				fprintf(fp, "%d", fw->fw_pts[i]);
+				fprintf(fp, "%u", fw->fw_pts[i]);
 		}
 
 	if (!(format & FMT_NONEWLINE))
 		putc('\n', fp);
 }
 
+void
+print_masq(FILE *fp, struct masq *ms, int format)
+{
+	unsigned long minutes, seconds, sec100s;
+	char *service;
+
+	if (format & FMT_HEADER) {
+		fputs("prot ", fp);
+		fprintf(fp, "%-8s ", "expire");
+		if (format & FMT_DELTAS) {
+			fprintf(fp, "%10s delta prevd ", "initseq");
+		}
+		fprintf(fp, "%-20s ", "source");
+		fprintf(fp, "%-20s ", "destination");
+		fputs("ports\n", fp);
+	}
+
+	switch (ms->kind) {
+	case IP_FW_F_TCP:
+		fprintf(fp, "%-5s", "tcp");
+		break;
+	case IP_FW_F_UDP:
+		fprintf(fp, "%-5s", "udp");
+		break;
+	}
+
+	sec100s = ms->expires % 100L;
+	seconds = (ms->expires / 100L) % 60;
+	minutes = ms->expires / 6000L;
+
+	fprintf(fp, "%02ld:%02ld.%02ld ", minutes, seconds, sec100s);
+
+	if (format & FMT_DELTAS) {
+		fprintf(fp, "%10lu %5hd %5hd ", (unsigned long) ms->initseq,
+			ms->delta, ms->pdelta);
+	}
+
+	if (format & FMT_NUMERIC) {
+		fprintf(fp, "%-20s ", addr_to_dotted(&(ms->src)));
+		fprintf(fp, "%-20s ", addr_to_dotted(&(ms->dst)));
+	} else {
+		fprintf(fp, "%-20s ", addr_to_anyname(&(ms->src)));
+		fprintf(fp, "%-20s ", addr_to_anyname(&(ms->dst)));
+	}
+
+	if (format & FMT_NUMERIC)
+		fprintf(fp, "%u (%u) -> %u", ms->sport, ms->mport, ms->dport);
+	else {
+		if ((service = port_to_service(ms->sport, ms->kind)) != NULL)
+			fprintf(fp, "%s (%u) -> ", service, ms->mport);
+		else
+			fprintf(fp, "%u (%u) -> ", ms->sport, ms->mport);
+		if ((service = port_to_service(ms->dport, ms->kind)) != NULL)
+			fputs(service, fp);
+		else
+			fprintf(fp, "%u", ms->dport);
+	}
+}
+
 int
 read_procinfo(FILE *fp, struct ip_fw *fwlist, int nfwlist)
 {
 	int i, n, nread = 0;
 	struct ip_fw *fw;
+	unsigned short tosand, tosxor;
+	unsigned long temp[5];
 
 	for (nread = 0; nread < nfwlist; nread++) {
 		fw = &fwlist[nread];
-		if ((n = fscanf(fp, "%lX/%lX->%lX/%lX %lX %hX %hu %hu %lu %lu",
-				&fw->fw_src.s_addr, &fw->fw_smsk.s_addr, 
-				&fw->fw_dst.s_addr, &fw->fw_dmsk.s_addr, 
-				&fw->fw_via.s_addr, &fw->fw_flg,
-				&fw->fw_nsp, &fw->fw_ndp,
+		if ((n = fscanf(fp, "%lX/%lX->%lX/%lX %16s %lX %hX %hu %hu %lu %lu",
+				&temp[0], &temp[1], &temp[2], &temp[3],
+				fw->fw_vianame, &temp[4],
+				&fw->fw_flg, &fw->fw_nsp, &fw->fw_ndp,
 				&fw->fw_pcnt, &fw->fw_bcnt)) == -1)
 			return nread;
-		else if (n != 10)
+		else if (n != 11)
 			exit_error(1, "unexpected input data");
-		else
+		else {
 			for (i = 0; i < IP_FW_MAX_PORTS; i++)
 				if (fscanf(fp, "%hu", &fw->fw_pts[i]) != 1)
 					exit_error(1, "unexpected input data");
+			if (fscanf(fp, " A%hX X%hX", &tosand, &tosxor) != 2)
+				exit_error(1, "unexpected input data");
+			if (!strcmp("-", fw->fw_vianame))
+				(fw->fw_vianame)[0] = '\0';
+			fw->fw_tosand = (unsigned char) tosand;
+			fw->fw_tosxor = (unsigned char) tosxor;
+		}
 
 		/* we always keep these addresses in network byte order */
-		fw->fw_src.s_addr = htonl(fw->fw_src.s_addr);
-		fw->fw_dst.s_addr = htonl(fw->fw_dst.s_addr);
-		fw->fw_via.s_addr = htonl(fw->fw_via.s_addr);
-		fw->fw_smsk.s_addr = htonl(fw->fw_smsk.s_addr);
-		fw->fw_dmsk.s_addr = htonl(fw->fw_dmsk.s_addr);
+		fw->fw_src.s_addr = (__u32) htonl(temp[0]);
+		fw->fw_dst.s_addr = (__u32) htonl(temp[2]);
+		fw->fw_via.s_addr = (__u32) htonl(temp[4]);
+		fw->fw_smsk.s_addr = (__u32) htonl(temp[1]);
+		fw->fw_dmsk.s_addr = (__u32) htonl(temp[3]);
+	}
+	return nread;
+}
+
+int
+read_masqinfo(FILE *fp, struct masq *mslist, int nmslist)
+{
+	int n, nread = 0;
+	struct masq *ms;
+	char buf[256];
+	unsigned long temp[3];
+
+	for (nread = 0; nread < nmslist; nread++) {
+		ms = &mslist[nread];
+		if ((n = fscanf(fp, " %s %lX:%hX %lX:%hX %hX %lX %hd %hd %lu",
+				buf, &temp[0], &ms->sport, &temp[1], &ms->dport,
+				&ms->mport, &temp[2], &ms->delta,
+				&ms->pdelta, &ms->expires)) == -1)
+			return nread;
+		else if (n != 10)
+			exit_error(1, "unexpected input data");
+
+		if (strcmp("TCP", buf) == 0)
+			ms->kind = IP_FW_F_TCP;
+		else if (strcmp("UDP", buf) == 0)
+			ms->kind = IP_FW_F_UDP;
+		else
+			exit_error(1, "unexpected input data");
+
+		/* we always keep these addresses in network byte order */
+		ms->src.s_addr = (__u32) htonl(temp[0]);
+		ms->dst.s_addr = (__u32) htonl(temp[1]);
+
+		ms->initseq = (__u32) temp[2];
 	}
 	return nread;
 }
@@ -1165,13 +1479,14 @@ fw_to_fwpkt(struct ip_fw *fw)
 	iph = &ipfwp.fwp_iph;
 
 	iph->version = IP_VERSION;
-	iph->ihl = sizeof(struct iphdr) / sizeof(unsigned long);
+	iph->ihl = sizeof(struct iphdr) / 4;
 	iph->frag_off &= htons(~IP_OFFSET);
 
 	iph->saddr = fw->fw_src.s_addr;
 	iph->daddr = fw->fw_dst.s_addr;
 
 	inaddrcpy(&ipfwp.fwp_via, &fw->fw_via);
+	strncpy(ipfwp.fwp_vianame, fw->fw_vianame, IFNAMSIZ);
 
 	switch (kind) {
 	case IP_FW_F_TCP:
@@ -1208,12 +1523,15 @@ do_setsockopt(int cmd, void *data, int length)
 	}
 	
 	ret = setsockopt(sockfd, IPPROTO_IP, cmd, (char *) data, length);
-	if (cmd != IP_FW_CHK_BLK && cmd != IP_FW_CHK_FWD) {
+	if (cmd != IP_FW_CHECK_IN && cmd != IP_FW_CHECK_OUT &&
+						cmd != IP_FW_CHECK_FWD) {
 		if (ret)
 			perror("ipfwadm: setsockopt failed");
 	} else {
 		if (!ret)
 			printf("packet accepted\n");
+		else if (errno == ECONNRESET)
+			printf("packet masqueraded\n");
 		else if (errno == ETIMEDOUT) {
 			printf("packet denied\n");
 			ret = 0;
@@ -1298,19 +1616,24 @@ exit_tryhelp(int status)
 void
 exit_printhelp()
 {
-	printf("Usage: %s -A command [options] (accounting)\n"
-		"       %s -B command [options] (blocking firewall)\n"
+	printf("%s\n\n"
+		"Usage: %s -A command [options] (accounting)\n"
 		"       %s -F command [options] (forwarding firewall)\n"
+		"       %s -I command [options] (input firewall)\n"
+		"       %s -O command [options] (output firewall)\n"
+		"       %s -M -l [options] (masquerading entries)\n"
 		"       %s -h (print this help information))\n\n",
-		program, program, program, program);
+		package_version, program, program, program, program,
+		program, program);
 
 	printf("Commands:\n"
-		"  -a [policy]	add rule (policy only required for firewall rules)\n"
-		"  -d [policy]	delete rule (policy only required for firewall rules)\n"
+		"  -i [policy]	insert rule (no policy for accounting rules)\n"
+		"  -a [policy]	append rule (no policy for accounting rules)\n"
+		"  -d [policy]	delete rule (no policy for accounting rules)\n"
 		"  -l		list all rules of this category\n"
 		"  -z		reset packet/byte counters of all rules of this category\n"
 		"  -f		remove all rules of this category\n"
-		"  -p policy	change default policy (either accept, deny, or reject)\n"
+		"  -p policy	change default policy (accept/masquerade/deny/reject)\n"
 		"  -c		check acceptance of IP packet\n\n"
 		"Options:\n"
 		"  -P		protocol (either tcp, udp, icmp, or all)\n"
@@ -1318,11 +1641,14 @@ exit_printhelp()
 		"		source specification\n"
 		"  -D address[/mask] [port ...]\n"
 		"		destination specification\n"
-		"  -I address	interface address\n\n"
-		"  -b		bidirectional mode\n"
+		"  -V address	interface address\n"
+		"  -W name	interface name\n"
+		"  -b		bidirectional match\n"
 		"  -e		extended output mode\n"
-		"  -k		turn on kernel logging for matching packets\n"
+		"  -k		match TCP packets with ACK set\n"
 		"  -n		numeric output of addresses and ports\n"
+		"  -o		turn on kernel logging for matching packets\n"
+		"  -t and xor	and/xor masks for TOS field\n"
 		"  -v		verbose mode\n"
 		"  -x		expand numbers (display exact values)\n"
 		"  -y		match TCP packets with SYN set and ACK cleared\n");
